package com.alterdekim.flash.decompiler.translator;

import com.alterdekim.flash.decompiler.action.ActionPush;
import com.alterdekim.flash.decompiler.item.*;
import com.alterdekim.flash.decompiler.action.ByteCodeAction;
import com.alterdekim.flash.decompiler.item.internal.InitObjectItem;
import com.alterdekim.flash.decompiler.tag.DoAction;
import com.alterdekim.flash.decompiler.util.ActionField;
import lombok.extern.slf4j.Slf4j;

import java.util.*;

@Slf4j
public class Flash2Java {

    private final List<String> pool;
    private final List<ByteCodeAction> actions;
    private final Stack<ByteCodeItem> stack; // ByteCodeAction
    private final Map<String, FlashVar> vars;

    public Flash2Java(DoAction doAction) {
        this.pool = doAction.getActionPool();
        this.stack = new Stack<>();
        this.actions = doAction.getActions();
        this.vars = new HashMap<>();
    }

    public Map<String, FlashVar> convert() {
        // abolish the activeness of variable inside of an HashMap. Use stack instead.
        for (ByteCodeAction action : actions) {
           // try {
                //log.info("Action type: " + action.getType().toString());
                switch (action.getType()) {
                    case DefineLocal:
                        defineLocal();
                        break;
                    case GetVariable:
                        getVariable();
                        break;
                    case SetMember:
                        setMember();
                        break;
                    case Add:
                        reduceViaAdder();
                        break;
                    case InitObject:
                        initObject();
                        break;
                    case Push:
                        this.stack.addAll(((ActionPush) action).getItems());
                        break;
                    case SetVariable:
                        setVariable();
                        break;
                    case GetMember:
                        getMember();
                        break;
                    default:
                        // ignore
                }
            //} catch (Exception e) {
                //log.error(e.getMessage());
            //}
        }
        return vars;
    }

    private void initObject() {
        Integer numberOfEntries = (Integer) this.stack.pop().getValue();
        HashMap<String, Object> obj = new HashMap<>();
        for( int i = 0; i < numberOfEntries; i++ ) {
            Object val = valConverter( this.stack.pop() );
            String key = (String) valConverter( this.stack.pop() );
            obj.put(key, val);
        }
        this.stack.push(new InitObjectItem(obj));
    }

    private void reduceViaAdder() {
        /*
        // get last 2 elements from stack
        List<Integer> si = new ArrayList<>();
        List<ByteCodeItem> g = new ArrayList<>();
        for( int i = this.stack.size()-1; i >= 0; i-- ) {
            ByteCodeAction e = this.stack.get(i);
            if( e.getType() != ActionField.Push ) continue;
            ActionPush push = (ActionPush) e;
            if( push.getItems().size() >= 2 ) {
                StringItem from = (StringItem) push.getItems().get(push.getItems().size()-1);
                StringItem to = (StringItem) push.getItems().get(push.getItems().size()-2);
                /*push.getItems().remove()
                ((String) to.getValue()) + ((String) from.getValue())*//*
            }
        }
        */
    }

    private void getMember() {
        Object val = this.stack.pop().getValue();
        String name = (String) valConverter( this.stack.pop() );
        log.info("SetMember: {}", name);
        String key = this.vars.keySet().stream().filter(k -> this.vars.get(k).isActive()).findFirst().get();
        HashMap<String, Object> g = (HashMap<String, Object>) this.vars.get(key).getVal();
        g.put(name, val);
        this.vars.put(key, new FlashVar(true, g));
    }

    private boolean hasActiveVariable() {
        return this.vars.keySet().stream().anyMatch(k -> this.vars.get(k).isActive());
    }

    private String getActiveVarKey() {
        return this.vars.keySet().stream().filter(k -> this.vars.get(k).isActive()).findFirst().get();
    }

    private void setMember() {
        ByteCodeItem val = this.stack.pop();

        if( this.stack.isEmpty() && this.hasActiveVariable() ) {
            String var = getActiveVarKey();

            return;
        }
        Integer index = (Integer) this.stack.pop().getValue();
        String var = this.vars.keySet().stream().filter(k -> this.vars.get(k).isActive()).findFirst().get();
        Map<String, Object> m = (Map<String, Object>) this.vars.get(var).getVal();
        m.put(String.valueOf(index), val.getValue());
        this.vars.put(var, new FlashVar(false, m)); // this may cause a bug (false in FlashVar arg)
    }

    private Object valConverter(ByteCodeItem item) {
        if ( item.getType() == ActionPushType.POOL_INDEX ) {
            return fromPool( (Integer) item.getValue() );
        } else if ( item.getType() == ActionPushType.STRING ) {
            return item.getValue();
        }
        return item.getValue();
    }

    private void setVariable() {
        Object val = this.stack.pop().getValue();
        String variableName = (String) valConverter( this.stack.pop() );
        vars.put(variableName, new FlashVar(false, val));
    }

    private void getVariable() {
        String variableName = (String) valConverter( this.stack.pop() );
        vars.keySet().forEach(k -> {
            FlashVar fv = vars.get(k);
            fv.setActive(false);
            vars.put(k, fv);
        });
        FlashVar v = vars.get(variableName);
        v.setActive(true);
        vars.put(variableName, v);
    }

    private void defineLocal() {
        ByteCodeItem val = this.stack.pop();
        String name = (String) valConverter( this.stack.pop() );
        // todo: implement assignment ability (from one value to another)
        vars.put(name,
                new FlashVar(false, val.getValue())
        );
    }

    private String fromPool(int index) {
        return pool.get(index);
    }
}
